@@ -17,7 +17,7 @@
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
  *
- * $Id: mmsh.c,v 1.5 2006/07/13 12:32:20 shawarma Exp $
+ * $Id: mmsh.c,v 1.16 2007/12/11 20:50:43 jwrdegoede Exp $
  *
  * MMS over HTTP protocol
  *   written by Thibaut Mattern
@@ -57,6 +57,9 @@
 # define lprintf(x...)
 #endif
 
+/* cheat a bit and call ourselves mms.c to keep the code in mmsio.h clean */
+#define __MMS_C__
+
 #include "bswap.h"
 #include "mmsh.h"
 #include "asfheader.h"
@@ -79,7 +82,7 @@
 #define CHUNK_TYPE_END         0x4524
 #define CHUNK_TYPE_ASF_HEADER  0x4824
 #define CHUNK_SIZE              65536  /* max chunk size */
-#define ASF_HEADER_SIZE          8192  /* max header size */
+#define ASF_HEADER_SIZE     (8192 * 2)  /* max header size */
 
 #define SCRATCH_SIZE             1024
 
@@ -162,6 +165,7 @@ struct mmsh_s {
   int           connect_port;
   char         *http_host;
   int           http_port;
+  int           http_request_number;
   char         *proxy_user;
   char         *proxy_password;
   char         *host_user;
@@ -177,7 +181,7 @@ struct mmsh_s {
   /* chunk */
   uint16_t      chunk_type;
   uint16_t      chunk_length;
-  uint16_t      chunk_seq_number;
+  uint32_t      chunk_seq_number;
   uint8_t       buf[CHUNK_SIZE];
 
   int           buf_size;
@@ -186,18 +190,21 @@ struct mmsh_s {
   uint8_t       asf_header[ASF_HEADER_SIZE];
   uint32_t      asf_header_len;
   uint32_t      asf_header_read;
-  int           seq_num;
   int           num_stream_ids;
   int           stream_ids[ASF_MAX_NUM_STREAMS];
   int           stream_types[ASF_MAX_NUM_STREAMS];
   uint32_t      packet_length;
   int64_t       file_length;
+  uint64_t      time_len; /* playback time in 100 nanosecs (10^-7) */
+  uint64_t      preroll;
+  uint64_t      asf_num_packets;
   char          guid[37];
   uint32_t      bitrates[ASF_MAX_NUM_STREAMS];
   uint32_t      bitrates_pos[ASF_MAX_NUM_STREAMS];
 
   int           has_audio;
   int           has_video;
+  int           seekable;
 
   off_t         current_pos;
   int           user_bandwidth;
@@ -425,10 +432,12 @@ static int get_answer (mms_io_t *io, mmsh_t *this) {
             if (strstr(features, "seekable")) {
               lprintf("seekable stream\n");
               this->stream_type = MMSH_SEEKABLE;
+              this->seekable = 1;
             } else {
               if (strstr(features, "broadcast")) {
                 lprintf("live stream\n");
                 this->stream_type = MMSH_LIVE;
+                this->seekable = 0;
               }
             }
           }
@@ -447,6 +456,7 @@ static int get_answer (mms_io_t *io, mmsh_t *this) {
   if (this->stream_type == MMSH_UNKNOWN) {
     lprintf ("mmsh: unknown stream type\n");
     this->stream_type = MMSH_SEEKABLE; /* FIXME ? */
+    this->seekable = 1;
   }
   return 1;
 }
@@ -462,7 +472,7 @@ static int get_chunk_header (mms_io_t *io, mmsh_t *this) {
   /* read chunk header */
   read_len = io_read(io, this->s, chunk_header, CHUNK_HEADER_LENGTH);
   if (read_len != CHUNK_HEADER_LENGTH) {
-    lprintf ("chunk header read failed, %d != %d\n", len, CHUNK_HEADER_LENGTH);
+    lprintf ("chunk header read failed, %d != %d\n", read_len, CHUNK_HEADER_LENGTH);
     return 0;
   }
   this->chunk_type       = LE_16 (&chunk_header[0]);
@@ -492,11 +502,14 @@ static int get_chunk_header (mms_io_t *io, mmsh_t *this) {
       return 0;
     }
   }
+  
+  if (this->chunk_type == CHUNK_TYPE_DATA || this->chunk_type == CHUNK_TYPE_END)
+    this->chunk_seq_number = LE_32 (&ext_header[0]);
+
   /* display debug infos */
 #ifdef DEBUG
   switch (this->chunk_type) {
     case CHUNK_TYPE_DATA:
-      this->chunk_seq_number = LE_32 (&ext_header[0]);
       lprintf ("chunk type:       CHUNK_TYPE_DATA\n");
       lprintf ("chunk length:     %d\n", this->chunk_length);
       lprintf ("chunk seq:        %d\n", this->chunk_seq_number);
@@ -505,7 +518,6 @@ static int get_chunk_header (mms_io_t *io, mmsh_t *this) {
       lprintf ("len2:             %d\n", LE_16(&ext_header[6]));
       break;
     case CHUNK_TYPE_END:
-      this->chunk_seq_number = LE_32 (&ext_header[0]);
       lprintf ("chunk type:       CHUNK_TYPE_END\n");
       lprintf ("continue: %d\n", this->chunk_seq_number);
       break;
@@ -571,6 +583,20 @@ static int get_header (mms_io_t *io, mmsh_t *this) {
     if (len != this->chunk_length) {
       return 0;
     } else {
+      /* check and 0 pad the first data chunk */
+      if (this->chunk_length > this->packet_length) {
+        lprintf ("mmsh: chunk_length(%d) > packet_length(%d)\n",
+                 this->chunk_length, this->packet_length);
+        return 0;
+      }
+
+      /* explicit padding with 0 */
+      if (this->chunk_length < this->packet_length)
+        memset(this->buf + this->chunk_length, 0,
+               this->packet_length - this->chunk_length);
+
+      this->buf_size = this->packet_length;
+
       return 1;
     }
   } else {
@@ -586,6 +612,8 @@ static void interp_header (mms_io_t *io, mmsh_t *this) {
   lprintf ("interp_header, header_len=%d\n", this->asf_header_len);
 
   this->packet_length = 0;
+  this->num_stream_ids = 0;
+  this->asf_num_packets = 0;
 
   /*
    * parse asf header
@@ -615,6 +643,9 @@ static void interp_header (mms_io_t *io, mmsh_t *this) {
           break;
         }
         this->file_length   = LE_64(this->asf_header + i + 40 - 24);
+        this->time_len       = LE_64(this->asf_header + i + 64 - 24);
+        //this->time_len       = LE_64(this->asf_header + i + 72 - 24);
+        this->preroll        = LE_64(this->asf_header + i + 80 - 24);
         lprintf ("file object, packet length = %d (%d)\n",
 		 this->packet_length, LE_32(this->asf_header + i + 96 - 24));
         break;
@@ -683,6 +714,11 @@ static void interp_header (mms_io_t *io, mmsh_t *this) {
         }
         break;
 
+      case GUID_ASF_DATA:
+        this->asf_num_packets = LE_64(this->asf_header + i + 40 - 24);
+        lprintf("mmsh: num_packets: %d\n", (int)this->asf_num_packets);
+        break;
+
       default:
         lprintf ("unknown object\n");
         break;
@@ -743,6 +779,8 @@ static int mmsh_tcp_connect(mms_io_t *io, mmsh_t *this) {
     progress += 1;
   } while ((res == MMS_IO_STATUS_TIMEOUT) && (progress < 30));
   if (res != MMS_IO_STATUS_READY) {
+    close (this->s);
+    this->s = -1;
     return 1;
   }
   lprintf ("connected\n");
@@ -750,7 +788,7 @@ static int mmsh_tcp_connect(mms_io_t *io, mmsh_t *this) {
   return 0;
 }
 
-static int mmsh_connect_int (mms_io_t *io, mmsh_t *this, int bandwidth) {
+static int mmsh_connect_int (mms_io_t *io, mmsh_t *this, off_t seek, uint32_t time_seek) {
   int    i;
   int    video_stream = -1;
   int    audio_stream = -1;
@@ -761,6 +799,15 @@ static int mmsh_connect_int (mms_io_t *io, mmsh_t *this, int bandwidth) {
   int    bandwitdh_left;
   char   stream_selection[10 * ASF_MAX_NUM_STREAMS]; /* 10 chars per stream */
   int    offset;
+  
+  /* Close exisiting connection (if any) and connect */
+  if (this->s != -1)
+    close(this->s);
+
+  if (mmsh_tcp_connect(io, this)) {
+    return 0;
+  }
+//  report_progress (stream, 30);
 
   /*
    * let the negotiations begin...
@@ -771,7 +818,7 @@ static int mmsh_connect_int (mms_io_t *io, mmsh_t *this, int bandwidth) {
   lprintf("first http request\n");
   
   snprintf (this->str, SCRATCH_SIZE, mmsh_FirstRequest, this->uri,
-            this->http_host, this->http_port, 1);
+            this->http_host, this->http_port, this->http_request_number++);
 
   if (!send_command (io, this, this->str))
     goto fail;
@@ -782,6 +829,8 @@ static int mmsh_connect_int (mms_io_t *io, mmsh_t *this, int bandwidth) {
     
   get_header(io, this);
   interp_header(io, this);
+  if (!this->packet_length || !this->num_stream_ids)
+    goto fail;
   
   close(this->s);
 //  report_progress (stream, 20);
@@ -804,11 +853,11 @@ static int mmsh_connect_int (mms_io_t *io, mmsh_t *this, int bandwidth) {
   }
 
   /* choose a video stream adapted to the user bandwidth */
-  bandwitdh_left = bandwidth - max_arate;
+  bandwitdh_left = this->user_bandwidth - max_arate;
   if (bandwitdh_left < 0) {
     bandwitdh_left = 0;
   }
-  lprintf("bandwitdh %d, left %d\n", bandwidth, bandwitdh_left);
+  lprintf("bandwitdh %d, left %d\n", this->user_bandwidth, bandwitdh_left);
 
   min_bw_left = bandwitdh_left;
   for (i = 0; i < this->num_stream_ids; i++) {
@@ -851,7 +900,7 @@ static int mmsh_connect_int (mms_io_t *io, mmsh_t *this, int bandwidth) {
   lprintf("second http request\n");
 
   if (mmsh_tcp_connect(io, this)) {
-    goto fail;
+    return 0;
   }
 
   /* stream selection string */
@@ -877,12 +926,14 @@ static int mmsh_connect_int (mms_io_t *io, mmsh_t *this, int bandwidth) {
   switch (this->stream_type) {
     case MMSH_SEEKABLE:
       snprintf (this->str, SCRATCH_SIZE, mmsh_SeekableRequest, this->uri,
-                this->http_host, this->http_port, 0, 0, 0, 2, 0,
+                this->http_host, this->http_port, time_seek,
+                (unsigned int)(seek >> 32),
+                (unsigned int)seek, this->http_request_number++, 0,
                 this->num_stream_ids, stream_selection);
       break;
     case MMSH_LIVE:
       snprintf (this->str, SCRATCH_SIZE, mmsh_LiveRequest, this->uri,
-                this->http_host, this->http_port, 2,
+                this->http_host, this->http_port, this->http_request_number++,
                 this->num_stream_ids, stream_selection);
       break;
   }
@@ -897,7 +948,10 @@ static int mmsh_connect_int (mms_io_t *io, mmsh_t *this, int bandwidth) {
 
   if (!get_header(io, this))
     goto fail;
+
   interp_header(io, this);
+  if (!this->packet_length || !this->num_stream_ids)
+    goto fail;
   
   for (i = 0; i < this->num_stream_ids; i++) {
     if ((this->stream_ids[i] != audio_stream) &&
@@ -915,6 +969,8 @@ static int mmsh_connect_int (mms_io_t *io, mmsh_t *this, int bandwidth) {
   }
   return 1;
 fail:
+  close(this->s);
+  this->s = -1;
   return 0;
 }
 
@@ -961,6 +1017,7 @@ mmsh_t *mmsh_connect (mms_io_t *io, void *data, const char *url, int bandwidth)
   this->has_video       = 0;
   this->current_pos     = 0;
   this->user_bandwidth  = bandwidth;
+  this->http_request_number = 1;
 
 //  report_progress (stream, 0);
 
@@ -984,6 +1041,7 @@ mmsh_t *mmsh_connect (mms_io_t *io, void *data, const char *url, int bandwidth)
     uri->port = MMSH_PORT;
   }
   if (this->proxy_url) {
+    char * uri_string;
     this->proto = (uri->scheme) ? strdup(uri->scheme) : NULL;
     this->connect_host = (proxy_uri->hostname) ? strdup(proxy_uri->hostname) : NULL;
     this->connect_port = proxy_uri->port;
@@ -994,7 +1052,7 @@ mmsh_t *mmsh_connect (mms_io_t *io, void *data, const char *url, int bandwidth)
     this->host_user = (uri->user) ? strdup(uri->user) : NULL;
     this->host_password = (uri->passwd) ? strdup(uri->passwd) : NULL;
     gnet_uri_set_scheme(uri,"http");
-    char * uri_string = gnet_uri_get_string(uri);
+    uri_string = gnet_uri_get_string(uri);
     this->uri = strdup(uri_string);
     g_free(uri_string);
   } else {
@@ -1007,7 +1065,22 @@ mmsh_t *mmsh_connect (mms_io_t *io, void *data, const char *url, int bandwidth)
     this->proxy_password = NULL;
     this->host_user =(uri->user) ?  strdup(uri->user) : NULL;
     this->host_password = (uri->passwd) ? strdup(uri->passwd) : NULL;
-    this->uri = (uri->path) ? strdup(uri->path) : NULL;
+
+    if (uri->path && uri->query) {
+
+      size_t plen = strlen(uri->path), qlen = strlen(uri->query);
+      this->uri = (char *)malloc(plen + qlen + 2);
+      memcpy(this->uri, uri->path, plen);
+      this->uri[plen] = '?';
+      memcpy(&this->uri[plen + 1], uri->query, qlen + 1);
+
+    } else {
+
+      char * uri_string = gnet_uri_get_string(uri);
+      this->uri = strdup(uri_string);
+      g_free(uri_string);
+
+	}
   }
   if (proxy_uri) {
     gnet_uri_delete(proxy_uri);
@@ -1021,13 +1094,8 @@ mmsh_t *mmsh_connect (mms_io_t *io, void *data, const char *url, int bandwidth)
     lprintf ("unsupported protocol\n");
     goto fail;
   }
-  
-  if (mmsh_tcp_connect(io, this)) {
-    goto fail;
-  }
-//  report_progress (stream, 30);
 
-  if (!mmsh_connect_int(io, this, this->user_bandwidth))
+  if (!mmsh_connect_int(io, this, 0, 0))
     goto fail;
 
 //  report_progress (stream, 100);
@@ -1076,7 +1144,7 @@ fail:
  * returned value:
  *  0: error
  *  1: data packet read
- *  2: new header read
+ *  2: new header and data packet read
  */
 static int get_media_packet (mms_io_t *io, mmsh_t *this) {
   int len = 0;
@@ -1093,13 +1161,15 @@ static int get_media_packet (mms_io_t *io, mmsh_t *this) {
 	if (this->chunk_seq_number == 0)
 	  return 0;
 
-	close (this->s);
-
-	if (mmsh_tcp_connect (io, this))
+        this->http_request_number = 1;
+	if (!mmsh_connect_int (io, this, 0, 0))
 	  return 0;
 
-	if (!mmsh_connect_int (io, this, this->user_bandwidth))
-	  return 0;
+	/* What todo with: current_pos ??
+	   Also our chunk_seq_numbers will probably restart from 0!
+	   If this happens with a seekable stream (does it ever?)
+	   and we get a seek request after this were fscked! */
+        this->seekable = 0;
 
 	/* mmsh_connect_int reads the first data packet */
 	/* this->buf_size is set by mmsh_connect_int */
@@ -1119,7 +1189,15 @@ static int get_media_packet (mms_io_t *io, mmsh_t *this) {
 	if (!get_header (io, this))
 	  return 0;
 	interp_header(io, this);
-	this->buf_size = this->packet_length;
+
+	/* What todo with: current_pos ??
+	   Also our chunk_seq_numbers might restart from 0!
+	   If this happens with a seekable stream (does it ever?) 
+	   and we get a seek request after this were fscked! */
+        this->seekable = 0;
+
+	/* get_header reads the first data packet */
+	/* this->buf_size is set by get_header */
 	return 2;
 
       default:
@@ -1180,6 +1258,10 @@ int mmsh_read (mms_io_t *io, mmsh_t *this, char *data, int len) {
 
   lprintf ("mmsh_read: len: %d\n", len);
 
+  /* Check if the stream didn't get closed because of previous errors */
+  if (this->s == -1)
+    return total;
+
   while (total < len) {
 
     if (this->asf_header_read < this->asf_header_len) {
@@ -1233,6 +1315,154 @@ int mmsh_read (mms_io_t *io, mmsh_t *this, char *data, int len) {
   return total;
 }
 
+off_t mmsh_seek (mms_io_t *io, mmsh_t *this, off_t offset, int origin) {
+  off_t dest;
+  off_t dest_packet_seq;
+  uint32_t orig_asf_header_len = this->asf_header_len;
+  uint32_t orig_asf_packet_len = this->packet_length;
+  
+  if (!this->seekable)
+    return this->current_pos;
+  
+  switch (origin) {
+    case SEEK_SET:
+      dest = offset;
+      break;
+    case SEEK_CUR:
+      dest = this->current_pos + offset;
+      break;
+    case SEEK_END:
+      dest = mmsh_get_length (this) + offset;
+    default:
+      return this->current_pos;
+  }
+
+  dest_packet_seq = dest - this->asf_header_len;
+  dest_packet_seq = dest_packet_seq >= 0 ?
+    dest_packet_seq / this->packet_length : -1;
+
+  if (dest_packet_seq < 0) {
+    if (this->chunk_seq_number > 0) {
+      lprintf("mmsh: seek within header, already read beyond first packet, resetting connection\n");
+      if (!mmsh_connect_int(io, this, 0, 0)) {
+        /* Oops no more connection let our caller know things are fscked up */
+        return this->current_pos = -1;
+      }
+      /* Some what simple / naive check to check for changed headers
+         if the header was changed we are once more fscked up */
+      if (this->asf_header_len != orig_asf_header_len ||
+          this->packet_length  != orig_asf_packet_len) {
+        lprintf("mmsh: AIIEEE asf header or packet length changed on re-open for seek\n");
+        /* Its a different stream, so its useless! */
+        close (this->s);
+        this->s = -1;
+        return this->current_pos = -1;
+      }
+    } else
+      lprintf("mmsh: seek within header, resetting buf_read\n");
+
+    // reset buf_read
+    this->buf_read = 0;
+    this->asf_header_read = dest;
+    return this->current_pos = dest;
+  }
+
+  // dest_packet_seq >= 0
+  if (this->asf_num_packets && dest == this->asf_header_len +
+      this->asf_num_packets*this->packet_length) {
+    // Requesting the packet beyond the last packet, can cause the server to
+    // not return any packet or any eos command.  This can cause
+    // mms_packet_seek() to hang.
+    // This is to allow seeking at end of stream, and avoid hanging.
+    --dest_packet_seq;
+    lprintf("mmsh: seek to eos!\n");
+  }
+
+  if (dest_packet_seq != this->chunk_seq_number) {
+
+    if (this->asf_num_packets && dest_packet_seq >= this->asf_num_packets) {
+      // Do not seek beyond the last packet.
+      return this->current_pos;
+    }
+    
+    lprintf("mmsh: seek to %d, packet: %d\n", (int)dest, (int)dest_packet_seq);
+    if (!mmsh_connect_int(io, this, (dest_packet_seq+1) * this->packet_length, 0)) {
+      /* Oops no more connection let our caller know things are fscked up */
+      return this->current_pos = -1;
+    }
+    /* Some what simple / naive check to check for changed headers
+       if the header was changed we are once more fscked up */
+    if (this->asf_header_len != orig_asf_header_len ||
+        this->packet_length  != orig_asf_packet_len) {
+      lprintf("mmsh: AIIEEE asf header or packet length changed on re-open for seek\n");
+      /* Its a different stream, so its useless! */
+      close (this->s);
+      this->s = -1;
+      return this->current_pos = -1;
+    }
+  }
+  else
+    lprintf("mmsh: seek within current packet, dest: %d, current pos: %d\n",
+       (int)dest, (int)this->current_pos);
+
+  /* make sure asf_header is seen as fully read by mmsh_read() this is needed
+     in case our caller tries to seek over part of the header, or when we've
+     done an actual packet seek as get_header() resets asf_header_read then. */
+  this->asf_header_read = this->asf_header_len;
+
+  /* check we got what we want */
+  if (dest_packet_seq == this->chunk_seq_number) {
+    this->buf_read = dest -
+      (this->asf_header_len + dest_packet_seq*this->packet_length);
+    this->current_pos = dest;
+  } else {
+    lprintf("Seek failed, wanted packet: %d, got packet: %d\n",
+      (int)dest_packet_seq, (int)this->chunk_seq_number);
+    this->buf_read = 0;
+    this->current_pos = this->asf_header_len + this->chunk_seq_number *
+      this->packet_length;
+  }
+
+  lprintf("current_pos after seek to %d: %d (buf_read %d)\n",
+    (int)dest, (int)this->current_pos, (int)this->buf_read);
+
+  return this->current_pos;
+}
+
+int mmsh_time_seek (mms_io_t *io, mmsh_t *this, double time_sec) {
+  uint32_t orig_asf_header_len = this->asf_header_len;
+  uint32_t orig_asf_packet_len = this->packet_length;
+
+  if (!this->seekable)
+    return 0;
+
+  lprintf("mmsh: time seek to %f secs\n", time_sec);
+  if (!mmsh_connect_int(io, this, 0, time_sec * 1000 + this->preroll)) {
+    /* Oops no more connection let our caller know things are fscked up */
+    this->current_pos = -1;
+    return 0;
+  }
+  /* Some what simple / naive check to check for changed headers
+     if the header was changed we are once more fscked up */
+  if (this->asf_header_len != orig_asf_header_len ||
+      this->packet_length  != orig_asf_packet_len) {
+    lprintf("mmsh: AIIEEE asf header or packet length changed on re-open for seek\n");
+    /* Its a different stream, so its useless! */
+    close (this->s);
+    this->s = -1;
+    this->current_pos = -1;
+    return 0;
+  }
+
+  this->asf_header_read = this->asf_header_len;
+  this->buf_read = 0;
+  this->current_pos = this->asf_header_len + this->chunk_seq_number *
+    this->packet_length;
+  
+  lprintf("mmsh, current_pos after time_seek:%d\n", (int)this->current_pos);
+
+  return 1;
+}
 
 void mmsh_close (mmsh_t *this) {
 
@@ -1266,9 +1496,37 @@ void mmsh_close (mmsh_t *this) {
 
 
 uint32_t mmsh_get_length (mmsh_t *this) {
-  return this->file_length;
+  /* we could / should return this->file_len here, but usually this->file_len
+     is longer then the calculation below, as usually an asf file contains an
+     asf index object after the data stream. However since we do not have a
+     (known) way to get to this index object through mms, we return a
+     calculated size of what we can get to when we know. */
+  if (this->asf_num_packets)
+    return this->asf_header_len + this->asf_num_packets*this->packet_length;
+  else
+    return this->file_length;
+}
+
+double mmsh_get_time_length (mmsh_t *this) {
+  return (double)(this->time_len) / 1e7;
+}
+
+uint64_t mmsh_get_raw_time_length (mmsh_t *this) {
+  return this->time_len;
 }
 
 off_t mmsh_get_current_pos (mmsh_t *this) {
   return this->current_pos;
 }
+
+uint32_t mmsh_get_asf_header_len (mmsh_t *this) {
+  return this->asf_header_len;
+}
+
+uint32_t mmsh_get_asf_packet_len (mmsh_t *this) {
+  return this->packet_length;
+}
+
+int mmsh_get_seekable (mmsh_t *this) {
+  return this->seekable;
+}
